"""
Copyright 2017 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import re
import logging
from collections import OrderedDict

# pylint: disable=import-error
import torch
import torch.nn as nn
from torch.autograd import Variable
# pylint: enable=import-error

import numpy

logger = logging.getLogger(__name__)

###############################################################################
class BYOM(nn.Module):				# pylint: disable=too-few-public-methods
	""" Bring-Your-Own-Module for PyTorch.

		This is just a simple PyTorch module that allows you to build up your
		module in pieces (using `setattr` to register parameters, and `func` to
		perform the forward pass). It gives you the benefits of a Module with
		the flexibility of Kur's graph interpretation.
	"""

	###########################################################################
	def __init__(self):
		""" Creates a new Module.
		"""
		super().__init__()
		self.func = None

	###########################################################################
	def forward(self, *inputs):
		""" Performs the forward pass.
		"""
		assert self.func is not None
		return self.func(self, *inputs)

###############################################################################
class Layer:
	""" Holds a PyTorch "layer". This is important because PyTorch copies the
		Modules for each GPU; that is, you cannot hold on to Module refernces,
		or even layer references, since they may change. This Layer class
		abstracts that away by dynamically grabbing the correct layer instance
		during the forward pass.
	"""

	###########################################################################
	def __init__(self, name, func=None):
		""" Creates a new layer.
		"""
		self.name = name
		self.func = func

	###########################################################################
	def __call__(self, module, *x):
		""" Grab the instantiated layer and evaluate it.
		"""
		operation = getattr(module, self.name)
		if self.func:
			return self.func(operation, *x)
		return operation(*x)

	###########################################################################
	@staticmethod
	def resolve(value):
		""" Convenience function for calling either Layers or standard PyTorch
			operations.
		"""
		if isinstance(value, Layer):
			return value
		elif hasattr(value, 'pure') and value.pure:
			return value
		return lambda _, *args: value(*args)

###############################################################################
class TorchModel:
	""" This holds the Torch graph and provides convenience functions for using
		it.
	"""

	###########################################################################
	def __init__(self, gpu=None):
		""" Creates a new model.
		"""
		self.model = BYOM()
		self.inputs = []
		self.outputs = None
		self.layer_map = {}
		self.gpu = gpu
		self._reuse = False
		self.final_model = None

	###########################################################################
	@property
	def allow_reuse(self):
		""" Getter for the `allow_reuse` property, which, if enabled, causes
			`add_layer` calls to reuse existing layer when possible, rather
			than raising an exception about duplicate names.
		"""
		return self._reuse

	###########################################################################
	@allow_reuse.setter
	def allow_reuse(self, value):
		self._reuse = value

	###########################################################################
	def set_outputs(self, outputs):
		""" Sets the model outputs.

			# Arguments

			outputs: an OrderedDict, or a list of `(name, layer)` tuples. Each
				`name` (or key of the OrderedDict) is the name of a layer, and
				each `layer` is a function as returned by `add_operation` or
				`add_layer`, once that function has been applied to other
				layers.
		"""
		self.outputs = OrderedDict(outputs)
		self.model.func = self.add_operation(bundle)(
			*self.outputs.values()
		)
		self.final_model = self.parallelize()

	###########################################################################
	def parallelize(self):
		""" Applies any parallelism requested.
		"""
		if not self.gpu:
			return self.model
		if isinstance(self.gpu, bool):
			devices = None
		else:
			devices = self.gpu
		return nn.DataParallel(self.model, devices).cuda()

	###########################################################################
	def to_torch(self, tensor):
		""" Creates a Torch tensor from an array.

			# Arguments

			tensor: one of tuple, list, numpy.ndarray, torch.Tensor
		"""
		#if self.gpu:
		#	if numpy_tensor.dtype.kind == 'f' and \
		#			numpy_tensor.dtype.itemsize != 4:
		#		numpy_tensor = numpy_tensor.astype(
		#			'{}f4'.format(numpy_tensor.dtype.byteorder)
		#		)
		if isinstance(tensor, (list, tuple)):
			tensor = numpy.array(tensor)
		if isinstance(tensor, numpy.ndarray):
			tensor = torch.from_numpy(tensor)
		tensor = tensor.float()
		if self.gpu:
			tensor = tensor.cuda()
		return tensor

	###########################################################################
	def predict(self, data):
		""" Performs the forward pass.
		"""
		inputs = tuple(
			Variable(self.to_torch(data[k]))
			for k in self.inputs
		)
		return self.final_model(*inputs)

	###########################################################################
	def cpu(self, x):
		""" Moves a tensor (or list/tuple of tensors) back to the CPU.
		"""
		if isinstance(x, (list, tuple)):
			return tuple(X.cpu() for X in x)
		return x.cpu()

	###########################################################################
	def test(self, data, loss):
		""" Performs a forward pass and calculates loss.

			# Arguments

			data: dict. A dictionary whose keys are the names of the model
				layers, and whose respective values are numpy arrays contain
				those values for the batch. For this function, both model
				inputs and outputs must be specified.
			loss: a list of loss functions, each one corresponding to the
				respective model output. These loss functions should take two
				arguments; the first is a list of the tensors needed to compute
				loss (e.g., the ground truth data), and the second is the model
				output to compare against.
		"""
		inputs = tuple(
			Variable(self.to_torch(data[k]))
			for k in self.inputs
		)
		predictions = self.final_model(*inputs)

		#######################################################################
		def get_loss(loss, prediction):
			""" Calculates the loss given a loss specification.
			"""
			loss_inputs = [x[1](None, *inputs) for x in loss[0]]
			return loss[1](loss_inputs, prediction)

		losses = [
			get_loss(loss[output], P)
			for output, P in zip(self.outputs, predictions)
		]

		return predictions, losses

	###########################################################################
	def move(self, module):
		""" Moves a module to the GPU, if the GPU is enabled.
		"""
		if self.gpu:
			return module.cuda()
		return module

	###########################################################################
	def placeholder(self, name, create=True):
		""" Creates a new input placeholder, or retrieves an existing one.
		"""

		# Placeholders are just named ways to access one of the input tensors.
		try:
			index = self.inputs.index(name)
		except ValueError:
			if not create:
				return None
			index = len(self.inputs)
			self.inputs.append(name)

		#######################################################################
		def calculate(_, *inputs):
			""" Applies the layer.
			"""
			if index >= len(inputs):
				raise IndexError('Out-of-range index: {}. Must be < {}. Note: '
					'we are trying to find "{}".'
					.format(index, len(inputs), calculate.target))
			return inputs[index]
		calculate.target = name
		calculate.index = index
		calculate.name = name

		return calculate

	###########################################################################
	@staticmethod
	def normalize_name(name):
		""" Creates a PyTorch-compatible layer name.

			In PyTorch, "layers" can be automatically tracked by a Module, but
			only if they are class attributes. So we will take a name, and get
			a Pythonically-allowed variable to use for the attribute.
		"""
		new_name = re.sub(r'^[^a-zA-Z_]|[^a-zA-Z0-9_]', r'_', name)
		new_name = 'layer_{}'.format(new_name)
		return new_name

	###########################################################################
	def backprop(self, losses):
		""" Runs the backward pass on the network.
		"""
		grads = [self.to_torch([1.0]) for _ in range(len(losses))]
		torch.autograd.backward(
			losses,
			grads
		)

	###########################################################################
	def add_operation(self, operation, name=None):# pylint: disable=no-self-use
		""" Adds a new operation to the graph.

			# Notes

			- Operations have no learnable parameters, and have no names.
		"""

		if name is None:
			name = operation.__name__

		#######################################################################
		def stack(*lower_layers):
			""" Returns a function suitable for using in a functional paradigm
				of model assemble.
			"""

			logger.debug('Connecting layers: %s feed into %s',
				[
					x.name if hasattr(x, 'name') else 'unknown'
					for x in lower_layers
				], name
			)

			###################################################################
			def calculate(module, *inputs):
				""" Applies the layer.
				"""
				result = Layer.resolve(operation)(
					module,
					*[x(module, *inputs) for x in lower_layers]
				)
				return result
			calculate.name = name
			calculate.op = operation

			return calculate

		stack.name = name

		return stack

	###########################################################################
	def add_variable(self, name, value, func=None):
		""" Adds a new variable.
		"""
		new_name = self.normalize_name(name)
		if name in self.layer_map:
			if not self.allow_reuse:
				raise ValueError('Duplicate name found: {}'.format(name))
		else:
			self.layer_map[name] = new_name
			logger.debug('Adding new layer: %s (internal: "%s")',
				name, new_name)
			setattr(self.model, new_name, value)

		return Layer(new_name, func)

	###########################################################################
	def add_layer(self, name, layer, func=None):
		""" Creates a new layer.

			# Notes

			- The layer must not already exist.
			- The layer will be registered as potentially containing learnable
			  parameters.
			- All learnable layers must be added using this function.
		"""
		layer = self.add_variable(name, layer, func)
		return self.add_operation(layer, name=name)

###############################################################################
def flatten(x):
	""" Flattens an input tensor.
	"""
	return x.view(x.size()[0], -1)

###############################################################################
def bundle(*x):
	""" Merge tensors.
	"""
	return x

###############################################################################
def swap_channels(x):
	""" Swaps the dimensions between Theano/PyTorch and TensorFlow dimension
		orderings.
	"""
	return torch.transpose(x, 1, x.dim()-1)

###############################################################################
def parallel(layer):
	""" Creates a parallel operation (i.e., map/distributed operation).
	"""
	def func(module, x):
		""" The actual wrapped operation.
		"""
		return torch.stack(
			tuple(Layer.resolve(layer)(module, X) for X in torch.unbind(x, 0)),
			0
		)
	func.pure = True
	return func

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2017 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import logging

from . import Optimizer, keras_clip, keras_wrap

logger = logging.getLogger(__name__)

###############################################################################
class SGD(Optimizer):
	""" Stochastic gradient descent optimizer
	"""

	###########################################################################
	def __init__(self, learning_rate=None, momentum=None, decay=None,
		nesterov=None, *args, **kwargs):
		""" Create a new Adam optimizer.

			# Arguments

			learning_rate: float. The learning rate to use.
			momentum: float. Momentum for parameter updates
			decay: float learning rate decay over each update
			nesterov: bool. Whether or not to apply Nesterov momentum.
		"""
		super().__init__(*args, **kwargs)

		self.learning_rate = learning_rate or 0.01
		self.momentum = momentum or 0.0
		self.decay = decay or 0.0
		self.nesterov = nesterov or False

		self.optimizer = None

	###########################################################################
	def get_optimizer(self, backend):
		""" Returns a backend-specific instantiation of the optimizer.
		"""
		if backend.get_name() == 'keras':
			import keras.optimizers as O		# pylint: disable=import-error
			self.optimizer = self.optimizer or O.SGD(
				lr=self.learning_rate,
				momentum=self.momentum,
				decay=self.decay,
				nesterov=self.nesterov,
				**keras_clip(self)
			)
			return keras_wrap(self.optimizer)
		elif backend.get_name() == 'pytorch':
			import torch.optim as optim			# pylint: disable=import-error
			if self.nesterov:
				logger.warning('The PyTorch backend does not use Nesterov '
					'momentum. Ignoring it...')
			if self.optimizer is None:
				self.optimizer = lambda params: optim.SGD(
					params,
					lr=self.learning_rate,
					momentum=self.momentum,
					weight_decay=self.decay
				)
			return self.optimizer
		else:
			raise ValueError('Unsupported backend "{}" for optimizer "{}"'
				.format(backend.get_name(), self.get_name()))

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2016 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from . import Layer

###############################################################################
class Flatten(Layer):					# pylint: disable=too-few-public-methods
	""" A layer which flattens the input tensor.
	"""

	###########################################################################
	def __init__(self, *args, **kwargs):
		""" Creates a new flattening layer.
		"""
		super().__init__(*args, **kwargs)

	###########################################################################
	def _build(self, model):
		""" Create the backend-specific placeholder.
		"""
		backend = model.get_backend()
		if backend.get_name() == 'keras':

			import keras.layers as L			# pylint: disable=import-error
			yield L.Flatten(
				name=self.name
			)

		elif backend.get_name() == 'pytorch':

			from kur.backend.pytorch.modules import flatten

			def connect(inputs):
				""" Connects the layer.
				"""
				assert len(inputs) == 1
				return {
					'shape' : self.shape([inputs[0]['shape']]),
					'layer' : model.data.add_operation(flatten)(
						inputs[0]['layer']
					)
				}

			yield connect

		else:
			raise ValueError(
				'Unknown or unsupported backend: {}'.format(backend))

	###########################################################################
	def shape(self, input_shapes):
		""" Returns the output shape of this layer for a given input shape.
		"""
		if len(input_shapes) > 1:
			raise ValueError('Flatten layers only take a single input.')
		input_shape = input_shapes[0]

		result = 1
		for x in input_shape:
			result *= x
		return (result, )

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2016 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from . import Layer, ParsingError

###############################################################################
class Convolution(Layer):				# pylint: disable=too-few-public-methods
	""" A vanilla convolution layer.

		# Properties

		kernels: int (required). The number of kernels (also called filters) to
			use.
		size: int or list of ints (required). The size of each kernel. If an
			integer is used, it is interpretted as a one-dimensional
			convolution, the same as if it were put into a length-1 list.
		strides: int or list of ints (optional; default: 1 in each dimension).
			The stride (subsampling rate) between convolutions. If a list, it
			must be the same length as `size` and specify the stride in each
			respective dimension. If a single number, it is used as the stride
			in each dimension.
		border: "valid" or "same" (default: "same").

		# Example

		```
		convolution:
		  kernels: 64
		  size: [2, 2]
		  strides: 1
		  border: relu
		```
	"""

	###########################################################################
	def __init__(self, *args, **kwargs):
		""" Creates a new convolution layer.
		"""
		super().__init__(*args, **kwargs)
		self.kernels = None
		self.size = None
		self.strides = None
		self.activation = None
		self.border = None

	###########################################################################
	def _parse(self, engine):
		""" Parses out the convolution layer.
		"""
		# Parse self
		if 'kernels' not in self.args:
			raise ParsingError('Missing key "kernels" in convolution '
				'container.')
		self.kernels = engine.evaluate(self.args['kernels'])
		try:
			self.kernels = int(self.kernels)
		except ValueError:
			raise ParsingError('"kernels" must evaluate to an integer.')

		if 'size' not in self.args:
			raise ParsingError('Missing key "size" in convolution container.')
		self.size = engine.evaluate(self.args['size'], recursive=True)
		if not isinstance(self.size, (list, tuple)):
			self.size = [self.size]
		if not 1 <= len(self.size) <= 3:
			raise ParsingError('Only convolutions with dimensions 1, 2, or 3 '
				'are supported.')
		for i in range(len(self.size)):
			try:
				self.size[i] = int(self.size[i])
			except ValueError:
				raise ParsingError('All "size" entries must evaluate to '
					'integers. We received this instead: {}'
					.format(self.size[i]))

		if 'strides' in self.args:
			self.strides = engine.evaluate(self.args['strides'], recursive=True)
			if not isinstance(self.strides, (list, tuple)):
				try:
					self.strides = int(self.strides)
				except ValueError:
					raise ParsingError('"strides" must evaluate to an integer '
						'or a list of integers.')
				self.strides = [self.strides] * len(self.size)
			else:
				if len(self.strides) != len(self.size):
					raise ParsingError('If "strides" is a list, it must be the '
						'same length as "size".')
				for i in range(len(self.strides)):
					try:
						self.strides[i] = int(self.strides[i])
					except ValueError:
						raise ParsingError('Each element of "strides" must '
							'evaluate to an integer.')
		else:
			self.strides = [1] * len(self.size)

		self.activation = None

		if 'border' in self.args:
			self.border = engine.evaluate(self.args['border'])
			if not isinstance(self.border, str) or \
				not self.border in ('valid', 'same'):
				raise ParsingError('"border" must be one of: "valid", "same".')
		else:
			self.border = 'same'

		assert self.border in ('same', 'valid')

	###########################################################################
	def _build(self, model):
		""" Instantiates the layer with the given backend.
		"""
		backend = model.get_backend()
		if backend.get_name() == 'keras':

			if backend.keras_version() == 1:
				import keras.layers as L			# pylint: disable=import-error

				kwargs = {
					'nb_filter' : self.kernels,
					'activation' : self.activation or 'linear',
					'border_mode' : self.border,
					'name' : self.name
				}

				if len(self.size) == 1:
					func = L.Convolution1D
					kwargs.update({
						'filter_length' : self.size[0],
						'subsample_length' : self.strides[0]
					})
				elif len(self.size) == 2:
					func = L.Convolution2D
					kwargs.update({
						'nb_row' : self.size[0],
						'nb_col' : self.size[1],
						'subsample' : self.strides
					})
				elif len(self.size) == 3:
					func = L.Convolution3D
					kwargs.update({
						'kernel_dim1' : self.size[0],
						'kernel_dim2' : self.size[1],
						'kernel_dim3' : self.size[2],
						'subsample' : self.strides
					})
				else:
					raise ValueError('Unhandled convolution dimension: {}. This '
						'is a bug.'.format(len(self.size)))

				yield func(**kwargs)

			else:

				import keras.layers.convolutional as L # pylint: disable=import-error

				kwargs = {
					'filters' : self.kernels,
					'activation' : self.activation or 'linear',
					'padding' : self.border,
					'name' : self.name,
					'kernel_size' : self.size,
					'strides' : self.strides
				}

				if len(self.size) == 1:
					func = L.Conv1D
				elif len(self.size) == 2:
					func = L.Conv2D
				elif len(self.size) == 3:
					func = L.Conv3D
				else:
					raise ValueError('Unhandled convolution dimension: {}. This '
						'is a bug.'.format(len(self.size)))

				if len(self.size) > 1:
					kwargs['data_format'] = 'channels_last'

				yield func(**kwargs)

		elif backend.get_name() == 'pytorch':

			# pylint: disable=import-error
			import torch.nn as nn
			# pylint: enable=import-error

			from kur.backend.pytorch.modules import swap_channels

			func = {
				1 : nn.Conv1d,
				2 : nn.Conv2d,
				3 : nn.Conv3d
			}.get(len(self.size))
			if not func:
				raise ValueError('Unhandled convolution dimension: {}. This '
					'is a bug.'.format(len(self.size)))

			if self.border == 'valid':
				padding = 0
			else:
				# "same" padding requires you to pad with P = S - 1 zeros
				# total. However, PyTorch always pads both sides of the input
				# tensor, implying that PyTorch only accepts padding P' such
				# that P = 2P'. This unfortunately means that if S is even,
				# then the desired padding P is odd, and so no P' exists.
				if any(s % 2 == 0 for s in self.size):
					raise ValueError('PyTorch convolutions cannot use "same" '
						'border mode when the receptive field "size" is even.')
				padding = tuple((s-1)//2 for s in self.size)

			layer = lambda in_channels: func(
				in_channels,
				self.kernels,
				tuple(self.size),
				stride=tuple(self.strides),
				padding=padding
			)

			def connect(inputs):
				""" Connects the layer.
				"""
				assert len(inputs) == 1
				output = model.data.add_operation(
					swap_channels
				)(inputs[0]['layer'])
				output = model.data.add_layer(
					self.name,
					layer(inputs[0]['shape'][-1])
				)(output)
				output = model.data.add_operation(
					swap_channels
				)(output)
				return {
					'shape' : self.shape([inputs[0]['shape']]),
					'layer' : output
				}

			yield connect

		else:
			raise ValueError(
				'Unknown or unsupported backend: {}'.format(backend))

	###########################################################################
	def shape(self, input_shapes):
		""" Returns the output shape of this layer for a given input shape.
		"""
		if len(input_shapes) > 1:
			raise ValueError('Convolutions only take a single input.')
		input_shape = input_shapes[0]
		if len(input_shape) != len(self.size) + 1:
			raise ValueError('Invalid input shape to a convolution layer: {}'
				.format(input_shape))

		def apply_border(input_shape, size):
			""" Resolves output shape differences caused by border mode.
			"""
			if self.border == 'same':
				return input_shape
			return input_shape - size + 1

		output_shape = tuple(
			(apply_border(input_shape[i], self.size[i]) + self.strides[i] - 1) \
				// self.strides[i] if input_shape[i] else None
			for i in range(len(self.size))
		) + (self.kernels, )
		return output_shape

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2017 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from . import OriginalSource

###############################################################################
class ChunkSource(OriginalSource):			# pylint: disable=abstract-method
	""" A chunk source is a source which can readily return different lengths
		of data at each iteration.
	"""

	DEFAULT_CHUNK_SIZE = 8192
	USE_BATCH_SIZE = object()
	ALL_ITEMS = object()

	###########################################################################
	@classmethod
	def default_chunk_size(cls):
		""" Returns the default chunk size for this source.
		"""
		return ChunkSource.DEFAULT_CHUNK_SIZE

	###########################################################################
	def __init__(self, chunk_size=None, *args, **kwargs):
		""" Creates a new ChunkSource.
		"""
		super().__init__(*args, **kwargs)
		self.requested_chunk_size = chunk_size or self.default_chunk_size()
		self.chunk_size = self.requested_chunk_size

	###########################################################################
	def set_chunk_size(self, chunk_size):
		""" Modifies the chunk size.
		"""
		self.chunk_size = chunk_size or self.default_chunk_size()

#### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2016 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import json
from . import Supplier
from ..sources import TextSource, TextLength, RawText


###############################################################################
class TextSupplier(Supplier):
    """ A supplier which parses out data from a JSON-Lines file (*.jsonl)
        the keys from the first JSON line object determine the source names, and
        all subsequent lines should have the same keys. The values should be strings
        which will be represented as one-hot encoding under a vocab.
    """

    ###########################################################################
    @classmethod
    def get_name(cls):
        """ Returns the name of the supplier.
        """
        return 'text'

    ###########################################################################
    def __init__(self, path, seq_len, vocabs, padding=None, pad_with=None, *args, **kwargs):
        """ Creates a new JSONL text dictionary supplier.

            # Arguments

            path: str. Filename of JSONL file (*.jsonl) to load.
            vocabs: dict. vocabs[key] = list of vocab items for this source
            padding: dict (optional0. padding[key] = 'left' or 'right' for 0 padding
        """
        super().__init__(*args, **kwargs)

        self.path = path
        self.seq_len = seq_len

        with open(self.path, 'r') as infile:
            keys = json.loads(next(infile)).keys()

        self.vocabs = vocabs

        if padding is None:
            self.padding = {
                k: 'right'
                for k in keys
            }
        else:
            self.padding = padding

        if pad_with is None:
            self.pad_with = {
                k: None
                for k in keys
            }
        else:
            self.pad_with = pad_with



        self.vocabs = {
            k: list(map(str, self.vocabs[k]))
            for k in self.vocabs
        }

        self.sources = None

        with open(self.path, 'r') as infile:
            self.num_entries = sum(1 for _ in infile if _.strip())

        self.sources = {}
        for k in keys:
            self.sources['raw_' + k] = RawText(
                self.path,
                k,
                self.num_entries
            )
            self.sources[k] = TextSource(
                'raw_' + k,
                self.vocabs[k],
                self.num_entries,
                self.seq_len,
                padding=self.padding[k],
                pad_with=self.pad_with[k]
            )
            self.sources[k + '_length'] = TextLength(
                'raw_' + k,
                self.num_entries
            )


    ###########################################################################
    def get_sources(self, sources=None):
        """ Returns all sources from this provider.
        """

        if sources is None:
            with open(self.path, 'r') as infile:
                keys = list(json.loads(next(infile)).keys())
                sources = [
                    k for k in keys
                ] + [
                    'raw_' + k for k in keys
                ] + [
                    k + '_length' for k in keys
                ]

        elif not isinstance(sources, (list, tuple)):
            sources = [sources]

        for source in sources:
            if source not in self.sources:
                raise KeyError(
                    'Invalid data key: {}. Valid keys are: {}'.format(
                        source, ', '.join(str(k) for k in self.sources.keys())
                ))

        return {k : self.sources[k] for k in sources}

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2016 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

__homepage__ = 'https://kur.deepgram.com'
from .version import __version__

from . import utils
from . import backend
from . import containers
from . import engine
from . import reader
from . import loss
from . import optimizer
from . import sources
from . import providers
from . import model
from . import supplier
from . import loggers
from .kurfile import Kurfile

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2016 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

###############################################################################
class Source:						# pylint: disable=too-few-public-methods
	""" Base class for all data sources.

		Data sources are responsible for producing data that will be presented
		to the model for training, validation, or prediction (evaluation).

		Each data source provides one of the input or output tensors that the
		network needs.
	"""

	###########################################################################
	def __init__(self):
		""" Create a new data source.
		"""
		pass

	###########################################################################
	def __iter__(self):
		""" Return an iterator to the data.

			The iterator should return a tensor of shape `(X, ) +
			self.shape()`, where `X` is the number of entries provided by this
			data source.

			# Note

			If `is_derived()` is False, then this should simply `yield` data.
			If `is_derived()` is True, then Kur will set up a bi-directional
			generator with two `yield`s expected. The first `yield` will give
			input to the Source, and the second `yield` is for returning the
			data back to Kur.
		"""
		raise NotImplementedError

	###########################################################################
	def __len__(self):
		""" Returns the total number of entries that this source can return, if
			known.

			# Return value

			If the total number of entries that this source can return is
			known, it should be returned. In the simplest case, this is the
			same as the number of training samples in an epoch. If this number
			is unknown or effectively infinite (because it is from some
			real-time data generator or is being streamed from some
			uncontrolled source, for example), then this should return <= 0.
		"""
		raise NotImplementedError

	###########################################################################
	def shape(self):
		""" Return the shape of the tensor (excluding batch size) returned by
			this data source.

			# Return value

			A shape tuple (length of each dimension in the tensor).
		"""
		raise NotImplementedError

	###########################################################################
	def on_added(self, provider):
		""" Called when this source is added to a provider.
		"""
		pass

	###########################################################################
	def is_derived(self):
		""" Returns True if this source is a derived source (requires other
			sources to be present in a provider in order to function).
		"""
		raise NotImplementedError

	###########################################################################
	def can_shuffle(self):
		""" Returns True if this data source can be shuffled.

			By default, sources are not shuffleable.
		"""
		raise False

	###########################################################################
	def shuffle(self, indices):
		""" Applies a permutation to the data.

			# Arguments

			indices: numpy array. List of indices to use in constructing the
				data permutation.

			# Return value

			None
		"""
		raise NotImplementedError

###############################################################################
class OriginalSource(Source):				# pylint: disable=abstract-method
	""" A source which is not derived.
	"""

	###########################################################################
	def is_derived(self):
		""" By definition, this source is not derived.
		"""
		return False

###############################################################################
class DerivedSource(Source):	# pylint: disable=abstract-method
	""" A source which is derived.
	"""

	###########################################################################
	def is_derived(self):
		""" By definition, this source is derived.
		"""
		return True

	############################################################################
	def requires(self):
		""" Returns a tuple of data sources that this source requires.
		"""
		raise NotImplementedError

	###########################################################################
	def __iter__(self):
		""" Derived sources take two steps: a `yield` to get input, and `yield`
			to send output. We provide a default implementation that is likely
			suitable for many derived sources.
		"""
		self.setup()
		while True:
			inputs = yield
			outputs = self.derive(inputs)
			yield outputs

	###########################################################################
	def __len__(self):
		""" Returns the total number of entries that this source can return.
		"""
		return 0

	###########################################################################
	def setup(self):
		""" Called at the beginning of iteration.
		"""
		pass

	###########################################################################
	def derive(self, inputs):
		""" Compute the derived source given its inputs.

			# Arguments

			inputs: list of arrays. Each list corresponds to the respective
				data source specified in `requires()`.

			# Return value

			An array that is passed back to the data provider.
		"""
		raise NotImplementedError

	###########################################################################
	def shuffle(self, indices):
		""" Applies a permutation to the data.
		"""
		pass

	###########################################################################
	def can_shuffle(self):
		""" Returns True if this data source can be shuffled.
		"""
		return True

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2017 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import logging
import numpy
from . import EvaluationHook

logger = logging.getLogger(__name__)

###############################################################################
class Transcript:
	""" Container for holding transcripts.
	"""

	###########################################################################
	def __init__(self, transcript):
		self.transcript = transcript

	###########################################################################
	def __str__(self):
		return self.transcript

	###########################################################################
	def __repr__(self):
		return '{}({})'.format(self.__class__.__name__, self.transcript)

###############################################################################
class TranscriptHook(EvaluationHook):
	""" Post-evaluation hook for speech data, which prints sample transcripts
	using argmax decoding
	"""

	###########################################################################
	@classmethod
	def get_name(cls):
		""" Returns the name of the evaluation hook.
		"""
		return 'transcript'

	###########################################################################
	@staticmethod
	def argmax_decode(output, rev_vocab, blank):
		""" output = matrix: timesteps * characters
		"""
		x = numpy.argmax(output, axis=1)
		tokens = []
		prev = None
		for c in x:
			if c == prev:
				continue
			if c != blank:
				tokens.append(c)
			prev = c
		offset = 1 if blank == 0 else 0
		return ''.join([rev_vocab[i-offset] for i in tokens])

	###########################################################################
	def __init__(self, warp=False, **kwargs):
		""" Creates a new transcript hook.
		"""

		super().__init__(**kwargs)
		self.warp = warp

	###########################################################################
	def apply(self, current, original, model=None):
		""" Applies the hook to the data.
		"""

		data, truth = current

		k = model.provider.keys.index('transcript_raw')
		vocab = model.provider.sources[k].vocab
		rev = {v : k for k, v in vocab.items()}

		blank = len(vocab) if not self.warp else 0

		prediction = data['asr'][0]
		result = {
			'prediction' : Transcript(self.argmax_decode(
				prediction,
				rev,
				blank
			)),
			'truth' : Transcript(''.join(
				rev.get(i, '') for i in truth['transcript_raw'][0]
			)) if truth is not None else None
		}
		print('Prediction: "{}"'.format(result['prediction']))
		print('Truth: "{}"'.format(result['truth']))

		return (result['prediction'], result['truth'])

#### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2016 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import logging

import numpy

from . import Layer, ParsingError

logger = logging.getLogger(__name__)

###############################################################################
class Recurrent(Layer):				# pylint: disable=too-few-public-methods
	""" A recurrent neural network.

		# Properties

		sequence: bool.
		size: int.
		bidirectional: bool.
		merge: one of (multiply, add, concat, average)
		type: one of (lstm, gru)

		# Example

		```
		recurrent:
		  size: 32
		  sequence: yes
		  bidirectional: yes
		  merge: average
		  type: lstm
		```
	"""

	MERGE_TYPES = ('multiply', 'add', 'concat', 'average')
	RNN_TYPES = ('lstm', 'gru')

	###########################################################################
	def __init__(self, *args, **kwargs):
		""" Creates a new recurrent layer.
		"""
		super().__init__(*args, **kwargs)
		self.type = None
		self.size = None
		self.sequence = None
		self.bidirectional = None
		self.merge = None

	###########################################################################
	def _parse(self, engine):
		""" Parses out the recurrent layer.
		"""
		self.sequence = engine.evaluate(self.args.get('sequence', True),
			recursive=True)
		if not isinstance(self.sequence, bool):
			raise ParsingError('Wrong type for "sequence" argument in '
				'recurrent layer. Expected bool, received: {}'
				.format(self.sequence))

		self.bidirectional = engine.evaluate(
			self.args.get('bidirectional', False),
			recursive=True
		)
		if not isinstance(self.bidirectional, bool):
			raise ParsingError('Wrong type for "bidirectional" argument in '
				'recurrent layer. Expected bool, received: {}'
				.format(self.bidirectional))

		self.merge = engine.evaluate(self.args.get('merge'), recursive=True)
		if not self.bidirectional:
			if self.merge is not None:
				raise ParsingError('Having a "merge" strategy in a '
					'"recurrent" layer only makes sense for bidirectional '
					'RNNs.')
		else:
			if self.merge is None:
				self.merge = 'average'
			if not isinstance(self.merge, str):
				raise ParsingError('Wrong type for "merge" argument in '
					'recurrent layer. Expected one of: {}. Received: {}'
						.format(', '.join(Recurrent.MERGE_TYPES), self.merge)
					)
			self.merge = self.merge.lower()
			if self.merge not in Recurrent.MERGE_TYPES:
				raise ParsingError('Bad value for "merge" argument in '
					'recurrent layer. Expected one of: {}. Received: {}'
						.format(', '.join(Recurrent.MERGE_TYPES), self.merge)
					)

		self.type = engine.evaluate(self.args.get('type', 'gru'),
			recursive=True)
		if not isinstance(self.type, str):
			raise ParsingError('Wrong type for "type" argument in recurrent '
				'layer. Expected one of: {}. Received: {}'.format(
					', '.join(Recurrent.RNN_TYPES), self.type
				))
		self.type = self.type.lower()
		if self.type not in Recurrent.RNN_TYPES:
			raise ParsingError('Bad value for "type" argument in recurrent '
				'layer. Expected one of: {}. Received: {}'.format(
					', '.join(Recurrent.RNN_TYPES), self.type
				))

		self.size = engine.evaluate(self.args.get('size'), recursive=True)
		if not isinstance(self.size, int):
			raise ParsingError('Bad or missing value for "size" argument in '
				'recurrent layer. Expected an integer. Received: {}'
				.format(self.size))

		if 'outer_activation' in self.args:
			self.activation = engine.evaluate(self.args['outer_activation'])
		else:
			self.activation = None

	###########################################################################
	def _build(self, model):
		""" Instantiates the layer with the given backend.
		"""
		backend = model.get_backend()
		if backend.get_name() == 'keras':
			if backend.keras_version() == 1:
				import keras.layers as L		# pylint: disable=import-error
			else:
				import keras.layers.recurrent as L # pylint: disable=import-error

			func = {
				'lstm' : L.LSTM,
				'gru' : L.GRU
			}.get(self.type)
			if func is None:
				raise ValueError('Unhandled RNN type: {}. This is a bug.'
					.format(self.type))

			if backend.keras_version() == 1:
				size_key = 'output_dim'
			else:
				size_key = 'units'

			kwargs = {
				'activation' : self.activation or 'relu',
				'return_sequences' : self.sequence,
				'go_backwards' : False,
				size_key : self.size
			}

			if self.bidirectional:
				kwargs['name'] = self.name + '_fwd'

				if self.merge in ('concat', ):
					if kwargs[size_key] % 2 != 0:
						logger.warning('Recurrent layer "%s" has an odd '
							'number for "size", but has a concat-type merge '
							'strategy. We are going to reduce its size by '
							'one.', self.name)
						kwargs[size_key] -= 1
					kwargs[size_key] //= 2

				forward = func(**kwargs)

				kwargs['go_backwards'] = True
				kwargs['name'] = self.name + '_rev'
				backward = func(**kwargs)

				def merge(tensor):
					""" Returns a bidirectional RNN.
					"""
					import keras.layers as L 	# pylint: disable=import-error
					if backend.keras_version() == 1:
						return L.merge(
							[forward(tensor), backward(tensor)],
							mode={
								'multiply' : 'mul',
								'add' : 'sum',
								'concat' : 'concat',
								'average' : 'ave'
							}.get(self.merge),
							name=self.name,
							**{
								'concat' : {'concat_axis' : -1}
							}.get(self.merge, {})
						)
					else:
						func = {
							'multiply' : L.multiply,
							'add' : L.add,
							'concat' : L.concatenate,
							'average' : L.average
						}.get(self.merge)
						return func(
							[forward(tensor), backward(tensor)],
							axis=-1,
							name=self.name
						)

				yield merge
			else:
				kwargs['name'] = self.name
				yield func(**kwargs)

		elif backend.get_name() == 'pytorch':

			# pylint: disable=import-error
			import torch.nn as nn
			# pylint: enable=import-error

			func = {
				'lstm' : nn.LSTM,
				'gru' : nn.GRU
			}.get(self.type)
			if func is None:
				raise ValueError('Unhandled RNN type: {}. This is a bug.'
					.format(self.type))

			if self.bidirectional and self.merge != 'concat':
				raise ValueError('PyTorch backend currently only supports '
					'"concat" mode for bidirectional RNNs.')

			if self.activation:
				raise ValueError('PyTorch backend currently only supports '
					'the default "outer_activation" value for RNNs.')

			def connect(inputs):
				""" Constructs the RNN layer.
				"""
				assert len(inputs) == 1
				size = self.size
				if self.bidirectional:
					if size % 2 != 0:
						logger.warning('Recurrent layer "%s" has an odd '
							'number for "size", but has a concat-type merge '
							'strategy. We are going to reduce its size by '
							'one.', self.name)
						size -= 1
					size //= 2

				kwargs = {
					'input_size' : inputs[0]['shape'][-1],
					'hidden_size' : size,
					'num_layers' : 1,
					'batch_first' : True,
					'bidirectional' : self.bidirectional,
					'bias' : True
				}

				def layer_func(layer, *inputs):
					""" Applies the RNN
					"""
					result, _ = layer(*(inputs + (None, )))
					if not self.sequence:
						return result[:, -1]
					return result

				return {
					'shape' : self.shape([inputs[0]['shape']]),
					'layer' : model.data.add_layer(
						self.name,
						func(**kwargs),
						func=layer_func
					)(inputs[0]['layer'])
				}

			yield connect

		else:
			raise ValueError(
				'Unknown or unsupported backend: {}'.format(backend))

	###########################################################################
	def shape(self, input_shapes):
		""" Returns the output shape of this layer for a given input shape.
		"""
		if len(input_shapes) > 1:
			raise ValueError('Recurrent layers only take a single input.')
		input_shape = input_shapes[0]
		if self.sequence:
			return input_shape[:-1] + (self.size, )
		return (self.size, )

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2017 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import copy

import numpy
import pytest

from kur.utils import neighbor_sort

###############################################################################
@pytest.fixture
def simple_data():
	return numpy.array([201, 202, 303, 101, 203, 102, 302, 103, 301])

###############################################################################
@pytest.fixture(
	params=[True, False]
)
def lotsa_data(request):
	with_repeats = request.param
	return numpy.random.choice(1000, size=500, replace=with_repeats)

###############################################################################
@pytest.fixture(
	params=[True, False]
)
def use_uniform(request):
	return request.param

###############################################################################
class TestNeighborSort:

	###########################################################################
	def test_even(self, simple_data, use_uniform):
		before = copy.deepcopy(simple_data)
		indices = neighbor_sort.argsort(
			simple_data,
			batch_size=3,
			neighborhood=5,
			uniform=use_uniform
		)

		# Make sure nothing changed.
		assert numpy.allclose(before, simple_data)

		# Do we have a valid permutation?
		assert indices.shape[0] == len(simple_data)
		assert indices.ndim == 1
		assert numpy.allclose(sorted(indices), numpy.arange(len(simple_data)))

		# Test the permutation.
		after = simple_data[indices]

		# Partition the result into three groups of three items each.
		groups = [after[3*i:(i+1)*3] for i in range(3)]
		# First sort by 100s.
		groups.sort(key=lambda x: x[0])
		# Now sort the items in each group.
		groups = [sorted(x) for x in groups]
		# Flatten the array.
		result = numpy.ravel(groups)

		# This should be fully sorted.
		assert numpy.allclose(result, sorted(before))

	###########################################################################
	def test_uniqueness(self, lotsa_data, use_uniform):
		indices = neighbor_sort.argsort(
			lotsa_data,
			batch_size=13,
			neighborhood=5,
			uniform=use_uniform
		)

		assert len(indices) == len(lotsa_data)
		assert indices.ndim == 1
		assert numpy.allclose(sorted(indices), numpy.arange(len(lotsa_data)))

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2016 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from . import Layer, ParsingError

###############################################################################
class Activation(Layer):				# pylint: disable=too-few-public-methods
	""" An activation layer.

		Some layers may include an 'activation' keyword. Those layers are
		intended to be equivalent to no-activation followed by this explicit
		activation layer.
	"""

	###########################################################################
	def __init__(self, *args, **kwargs):
		""" Creates a new activation layer.
		"""
		super().__init__(*args, **kwargs)
		self.type = None

	###########################################################################
	def _parse(self, engine):
		""" Parse the layer.
		"""
		self.type = self.args

	###########################################################################
	def _build(self, model):
		""" Create the backend-specific placeholder.
		"""
		backend = model.get_backend()
		if backend.get_name() == 'keras':

			import keras.layers as L			# pylint: disable=import-error
			yield L.Activation(
				'linear' if self.type == 'none' or self.type is None \
					else self.type,
				name=self.name
			)

		elif backend.get_name() == 'pytorch':

			import torch.nn.functional as F		# pylint: disable=import-error
			func = {
				'relu' : F.relu,
				'tanh' : F.tanh,
				'sigmoid' : F.sigmoid,
				'softmax' : F.log_softmax
			}.get(self.type.lower())
			if func is None:
				raise ValueError('Unsupported activation function "{}" for '
					'backend "{}".'.format(self.type, backend.get_name()))

			def connect(inputs):
				""" Connects the layer.
				"""
				assert len(inputs) == 1
				return {
					'shape' : self.shape([inputs[0]['shape']]),
					'layer' : model.data.add_operation(func)(
						inputs[0]['layer']
					)
				}

			yield connect

		else:
			raise ValueError(
				'Unknown or unsupported backend: {}'.format(backend))

	###########################################################################
	def shape(self, input_shapes):
		""" Returns the output shape of this layer for a given input shape.
		"""
		if len(input_shapes) > 1:
			raise ValueError('Activations only take a single input.')
		input_shape = input_shapes[0]
		return input_shape

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2016 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import logging
import numpy
from . import ShuffleProvider
from ..sources import ChunkSource
from ..utils import neighbor_sort

logger = logging.getLogger(__name__)

###############################################################################
class BatchProvider(ShuffleProvider): # pylint: disable=too-few-public-methods
	""" A provider which tries to return the same number of samples at every
		batch.

		All batches except for the last batch are guaranteed to be constant.
		The last batch might be different if there are not enough samples
		available to populate a complete batch.
	"""

	DEFAULT_BATCH_SIZE = 32

	###########################################################################
	@classmethod
	def get_name(cls):
		""" Returns the name of the provider.
		"""
		return 'batch_provider'

	###########################################################################
	def __init__(self, batch_size=None, num_batches=None,
		force_batch_size=False, neighborhood_sort=None, neighborhood_size=None,
		neighborhood_growth=None, *args, **kwargs):
		""" Creates a new batch provider.

			# Arguments

			batch_size: int. The number of batches to return each iteration.
		"""
		super().__init__(*args, **kwargs)

		self.batch_size = batch_size or BatchProvider.DEFAULT_BATCH_SIZE
		logger.debug('Batch size set to: %d', self.batch_size)

		self.num_batches = num_batches

		self.force = force_batch_size
		if self.force:
			logger.debug('Batch provider will force batches of exactly %d '
				'samples.', self.batch_size)

		if neighborhood_sort:
			if self.keys is None:
				raise ValueError('Cannot use "neighborhood_sort" with unnamed '
					'sources.')
			try:
				neighborhood_data = self.sources[self.keys.index(neighborhood_sort)]
			except ValueError:
				raise ValueError('Could not find the "neighborhood_sort" key '
					'"{}" in list of available sources: {}'
					.format(neighborhood_sort, ', '.join(self.keys)))

			if len(neighborhood_data) <= 0:
				raise ValueError('Data sorting requires a finite source.')
		else:
			neighborhood_data = None

		self.neighborhood_data = neighborhood_data
		self.neighborhood_key = neighborhood_sort
		self.neighborhood_size = neighborhood_size
		self.neighborhood_growth = neighborhood_growth

	###########################################################################
	@property
	def num_batches(self):
		""" Returns the number of batches.
		"""
		return self._num_batches

	###########################################################################
	@num_batches.setter
	def num_batches(self, value):
		""" Sets the number of batches this provider should use per epoch.
		"""
		self._num_batches = value # pylint: disable=attribute-defined-outside-init
		if self.num_batches is not None:
			logger.debug('Maximum number of batches set to: %d',
				self.num_batches)
			if self.entries > 0:
				self.entries = min( # pylint: disable=attribute-defined-outside-init
					self.entries,
					self.num_batches * self.batch_size
				)

	###########################################################################
	def add_source(self, source, name=None):
		""" Adds a new data source to an existing provider.
		"""
		super().add_source(source, name=name)
		if self.num_batches is not None:
			if self.entries > 0:
				self.entries = min( # pylint: disable=attribute-defined-outside-init
					self.entries,
					self.num_batches * self.batch_size
				)

	###########################################################################
	def __iter__(self):
		""" Retrieves the next batch of data.

			# Return value

			A tensor of shape `(X, ) + self.shape()`, where `X` is the number
			of entries provided by this data source.
		"""

		# Should always be the first call in __iter__
		self.pre_iter()

		for source in self.sources:
			if isinstance(source, ChunkSource):
				if source.requested_chunk_size is ChunkSource.USE_BATCH_SIZE:
					source.set_chunk_size(self.batch_size)

		logger.debug('Preparing next batch of data...')

		iterators = [iter(source) for source in self.sources]
		ordering = self.order_sources()
		dependencies = self.source_dependencies()
		queues = [[] for it in iterators]
		sentinel = object()
		proceed = True
		batches_produced = 0

		for it, source in zip(iterators, self.sources):
			if source.is_derived():
				next(it)

		while iterators and proceed:

			if batches_produced:
				logger.debug('Preparing next batch of data...')

			result = [None for it in iterators]

			# Go over each data source.
			for i in ordering:
				it = iterators[i]
				source = self.sources[i]
				depends = dependencies[i]

				if source.is_derived():
					requirements = [result[k] for k in depends]
					if any(x is None for x in requirements):
						raise ValueError('One of the dependent data sources '
							'has not been calculated yet. This is a bug.')
					try:
						result[i] = it.send(requirements)
						next(it)
					except StopIteration:
						proceed = False

				else:
					# Get enough data out of each one.
					while len(queues[i]) < self.batch_size:

						# Get the next batch. If there isn't any data left,
						# flag it.  After all, we may have collected at least
						# some data during the `while` loop, and we should
						# return that data.
						x = next(it, sentinel)
						if x is sentinel:
							proceed = False
							break

						# Add the data to the queue.
						if not queues[i]:
							queues[i] = x
						else:
							queues[i].extend(x)

					# Get the data ready.
					result[i] = numpy.array(queues[i][:self.batch_size])
					queues[i] = queues[i][self.batch_size:]

					if not result[i].shape[0]:
						logger.debug('An original source has no data.')
						return

			logger.debug('Next batch of data has been prepared.')

			lens = {len(q) for q in result}
			if len(lens) == 1:
				the_size = lens.pop()
				if self.force and the_size != self.batch_size:
					break
				elif the_size:
					yield self.wrap(result)

					batches_produced += 1
					if self.num_batches is not None:
						if batches_produced >= self.num_batches:
							break
				else:
					break
			else:
				if proceed:
					logger.error('Unequal length batches were produced: '
						'%s', self.wrap(result))
					raise ValueError('Managed to accumulate unequal-length '
						'batches, but we were still told to continue '
						'iterating. This is a bug.')

				smallest = min(lens)
				if smallest == 0:
					break
				if self.force and smallest != self.batch_size:
					break
				result = [x[:smallest] for x in result]
				yield self.wrap(result)

	###########################################################################
	def pre_iter(self):
		""" Pre-iteration hook.
		"""

		if not self.randomize or \
				self.shuffle_after or \
				not self.neighborhood_key:
			super().pre_iter()
			return

		logger.info('Calculating the nearest-neighbor shuffle indices using '
			'data source "%s".', self.neighborhood_key)

		# Create a local, complete copy of the sort-by data.
		# TODO: Can this safely be cached?
		data = numpy.empty(
			(len(self.neighborhood_data), ) + self.neighborhood_data.shape()
		)
		start = 0
		for batch in self.neighborhood_data:
			data[start:start+len(batch)] = batch[:]
			start += len(batch)

		# Determine the sort indices.
		indices = neighbor_sort.argsort(
			data,
			self.batch_size,
			self.neighborhood_size,
			self.neighborhood_growth
		)

		# Apply the indices.
		for source in self.sources:
			source.shuffle(indices)

		logger.debug('Finished shuffling.')

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2016 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import numpy as np
import json
from vocab import *


# this script shows sections of size `window` from the ends
# of each data file. its output looks like this:
# peek at train:
# ------
# "the project gutenberg ebook of" --> " "
# "he project gutenberg ebook of " --> "p"
# "e project gutenberg ebook of p" --> "r"
# ------
# "o particular resentment by his" --> " "
# " particular resentment by his " --> "h"
# "particular resentment by his h" --> "a"


# peek at validate:
# ------
# "articular resentment by his ha" --> "v"
# "rticular resentment by his hav" --> "i"
# "ticular resentment by his havi" --> "n"
# ------
# "ingley for a kingdom upon my h" --> "o"
# "ngley for a kingdom upon my ho" --> "n"
# "gley for a kingdom upon my hon" --> "o"


# peek at test:
# ------
# "ley for a kingdom upon my hono" --> "u"
# "ey for a kingdom upon my honou" --> "r"
# "y for a kingdom upon my honour" --> " "
# ------
# "sting your time with me. mr. b" --> "i"
# "ting your time with me. mr. bi" --> "n"
# "ing your time with me. mr. bin" --> "g"


# peek at evaluate:
# ------
# "ng your time with me. mr. bing" --> "l"
# "g your time with me. mr. bingl" --> "e"
# " your time with me. mr. bingle" --> "y"
# ------
# "they had yet learnt to care fo" --> "r"
# "hey had yet learnt to care for" --> " "
# "ey had yet learnt to care for " --> "a"


window = 3

def get_data(name):
    with open('data/%s.jsonl' % name, 'r') as infile:
        data = {}

        data = {
            key: [value]
            for key, value in json.loads(next(infile)).items()
        }
        for line in infile:
            line = line.strip()
            if not line:
                continue
            for key, value in json.loads(line).items():
                data[key].append(value)

        for key in data:
            data[key] = np.array(data[key])

    return data


def view_data(name):
    data = get_data(name)

    print('\n\npeek at %s:' % name)
    for j_range in (range(window), range(len(data['in_seq']) - window, len(data['in_seq'])),):
        print('------')
        for j in j_range:
            print(
                '"%s" --> "%s"' % (
                    ''.join([
                        int_to_char[np.argmax(_)]
                        for _ in data['in_seq'][j]
                    ]),
                    int_to_char[np.argmax(data['out_char'][j])]
                )
            )

if __name__ == '__main__':
    for name in ('train', 'validate', 'test', 'evaluate',):
        view_data(name)



"""
Copyright 2017 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import os
import logging

import numpy

from . import TrainingHook
from ...loggers import PersistentLogger, Statistic

logger = logging.getLogger(__name__)

###############################################################################
class PlotHook(TrainingHook):
	""" Hook for creating plots of loss.
	"""

	###########################################################################
	@classmethod
	def get_name(cls):
		""" Returns the name of the hook.
		"""
		return 'plot'

	###########################################################################
	def __init__(self, loss_per_batch=None, loss_per_time=None,
		throughput_per_time=None, *args, **kwargs):
		""" Creates a new plotting hook.
		"""
		super().__init__(*args, **kwargs)

		plots = dict(zip(
			('loss_per_batch', 'loss_per_time', 'throughput_per_time'),
			(loss_per_batch, loss_per_time, throughput_per_time)
		))

		self.plots = plots
		for k, v in self.plots.items():
			if v is not None:
				self.plots[k] = os.path.expanduser(os.path.expandvars(v))

		try:
			import matplotlib					# pylint: disable=import-error
		except:
			logger.exception('Failed to import "matplotlib". Make sure it is '
				'installed, and if you have continued trouble, please check '
				'out our troubleshooting page: https://kur.deepgram.com/'
				'troubleshooting.html#plotting')
			raise

		matplotlib.use('Agg')

	###########################################################################
	def notify(self, status, log=None, info=None):
		""" Creates the plot.
		"""
		from matplotlib import pyplot as plt	# pylint: disable=import-error

		logger.debug('Plotting hook received training message.')

		if status not in (
			TrainingHook.TRAINING_END,
			TrainingHook.VALIDATION_END,
			TrainingHook.EPOCH_END
		):
			logger.debug('Plotting hook does not handle this status.')
			return

		if log is None:
			logger.warning('Plot hook was requested, but no logger is '
				'available.')
			return
		if not isinstance(log, PersistentLogger):
			logger.warning('Plot hook was requested, but the logger does not '
				'track statistics.')
			return

		log.flush()

		# Load the data
		batch, time, loss = log.load_statistic(
			Statistic(Statistic.Type.BATCH, 'loss', 'total')
		)
		if loss is None:
			batch, time, loss = log.load_statistic(
				Statistic(Statistic.Type.TRAINING, 'loss', 'total')
			)
			if loss is None:
				logger.debug('No training data available for plotting yet.')
				return
			logger.debug('Using per-epoch training statistics for plotting.')
		else:
			logger.debug('Using per-batch training statistics for plotting.')

		vbatch, vtime, vloss = log.load_statistic(
			Statistic(Statistic.Type.VALIDATION, 'loss', 'total')
		)

		path = self.plots.get('loss_per_batch')
		if path:
			plt.xlabel('Batch')
			plt.ylabel('Loss')

			if batch is None:
				batch = numpy.arange(1, len(loss)+1)
			t_line, = plt.plot(batch, loss, 'co-', label='Training Loss')

			if vbatch is None:
				vbatch = numpy.arange(1, len(vloss)+1)
			v_line, = plt.plot(vbatch, vloss, 'mo-', label='Validation Loss')

			plt.legend(handles=[t_line, v_line])
			plt.savefig(path, transparent=True, bbox_inches='tight')
			plt.clf()

			logger.debug('Loss-per-batch plot saved to: %s', path)

		path = self.plots.get('loss_per_time')
		if path:
			plt.xlabel('Time')
			plt.ylabel('Loss')

			if time is None:
				logger.warning('No training timestamps available for '
					'loss-per-time plot.')
				t_line = None
			else:
				t_line, = plt.plot(time, loss, 'co-', label='Training Loss')

			if vtime is None:
				logger.warning('No validation timestamps available for '
					'loss-per-time plot.')
				v_line = None
			else:
				v_line, = plt.plot(vtime, vloss, 'mo-',
					label='Validation Loss')

			plt.legend(handles=[x for x in (t_line, v_line) if x is not None])
			plt.savefig(path, transparent=True, bbox_inches='tight')
			plt.clf()

			logger.debug('Loss-per-time plot saved to: %s', path)

		path = self.plots.get('throughput_per_time')
		if path:
			if time is None or batch is None:
				logger.warning('No training timestamps or batches available '
					'for throughput-per-time plot.')
			else:
				plt.xlabel('Time')
				plt.ylabel('Throughput (Batches/Second)')

				throughput = numpy.diff(batch) / numpy.diff(time)
				t_line, = plt.plot(time[1:], throughput, 'co-',
					label='Training Throughput')

				plt.legend(handles=(t_line, ))
				plt.savefig(path, transparent=True, bbox_inches='tight')
				plt.clf()

				logger.debug('Throughput-per-time plot saved to: %s', path)

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2017 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from . import Layer, ParsingError

###############################################################################
class Embedding(Layer):				# pylint: disable=too-few-public-methods
	""" A layer maps integer labels to a higher-dimensional embedding.
	"""

	###########################################################################
	def __init__(self, *args, **kwargs):
		""" Creates a new embedding layer.
		"""
		super().__init__(*args, **kwargs)
		self.vocab_size = None
		self.size = None

	###########################################################################
	def _parse(self, engine):
		""" Parse the layer.
		"""
		if not isinstance(self.args, dict):
			raise ParsingError('Embedding layer expected a dictionary of '
				'parameters, but instead received: {}'.format(self.args))

		if 'vocab_size' not in self.args:
			raise ParsingError('Missing "vocab_size" key in Embedding layer.')
		self.vocab_size = engine.evaluate(self.args['vocab_size'],
			recursive=True)
		try:
			self.vocab_size = int(self.vocab_size)
		except ValueError:
			raise ParsingError('Key "vocab_size" in Embedding layer must be '
				'an integer.')
		if self.vocab_size < 1:
			raise ParsingError('Key "vocab_size" in Embedding layer must be '
				'>= 1.')

		if 'size' not in self.args:
			raise ParsingError('Missing "size" key in Embedding layer.')
		self.size = engine.evaluate(self.args['size'], recursive=True)
		try:
			self.size = int(self.size)
		except ValueError:
			raise ParsingError('Key "size" in Embedding layer must be an '
				'integer.')
		if self.size < 1:
			raise ParsingError('Key "size" in Embedding layer must be >= 1.')

	###########################################################################
	def _build(self, model):
		""" Create the backend-specific placeholder.
		"""
		backend = model.get_backend()
		if backend.get_name() == 'keras':

			import keras.layers as L			# pylint: disable=import-error

			yield L.Embedding(
				self.vocab_size,
				self.size,
				name=self.name
			)

		elif backend.get_name() == 'pytorch':

			import torch.nn as nn				# pylint: disable=import-error

			def connect(inputs):
				""" Connects the layer.
				"""
				assert len(inputs) == 1
				return {
					'shape' : self.shape([inputs[0]['shape']]),
					'layer' : model.data.add_layer(
						self.name,
						lambda *x: nn.Embedding(
							self.vocab_size,
							self.size
						)(*[X.long() for X in x])
					)(inputs[0]['layer'])
				}

			yield connect

		else:
			raise ValueError(
				'Unknown or unsupported backend: {}'.format(backend))

	###########################################################################
	def shape(self, input_shapes):
		""" Returns the output shape of this layer for a given input shape.
		"""
		if len(input_shapes) > 1:
			raise ValueError('Embedding layers only take a single input.')
		input_shape = input_shapes[0]
		if len(input_shape) != 1:
			raise ValueError('Embedding layers only accept flat (1D) inputs.')
		return input_shape + (self.size, )

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2016 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import logging

import numpy

from . import Loss
from ..sources import RepeatSource, DerivedSource
from ..engine import PassthroughEngine
from ..utils import can_import

logger = logging.getLogger(__name__)

###############################################################################
class ScaledSource(DerivedSource):
	""" Derived source which scales `input_length` by the same
		amount that the length of `scale_this` is scaled down
		to `target`.
	"""
	###########################################################################
	def __init__(self, model, relative_to, to_this, scale_this):
		super().__init__()
		self.model = model
		self.relative_to = relative_to
		self.to_this = to_this
		self.scale_this = scale_this
		self.normal_shape = None

	###############################################################
	def derive(self, inputs):
		# Break it apart
		sizes, = inputs
		if sizes.ndim < 2:
			sizes = numpy.expand_dims(sizes, -1)
		outputs = numpy.array(
			[
				self.model.get_shape_at_layer(
					name=self.to_this,
					assumptions={
						self.relative_to : \
							(x[0], ) + tuple(self.normal_shape[1:])
					}
				)[0]
				for x in sizes
			],
			dtype='int32'
		)
		return numpy.expand_dims(outputs, axis=1)

	###############################################################
	def setup(self):
		self.normal_shape = self.model.get_shape_at_layer(
			name=self.relative_to
		)

	###############################################################
	def shape(self):
		return (None, 1)

	###############################################################
	def requires(self):
		return (self.scale_this, )

###################################################################
class FlattenSource(DerivedSource):
	""" Derived source which converts a rectangular array of
		labels (one for each sample, and all padded to the same
		width), along with the label lengths (the number of non-
		padded entries in each sample) and produces a single, long,
		flattened list of labels.
	"""

	###############################################################
	def __init__(self, label, label_length):
		super().__init__()
		self.label = label
		self.label_length = label_length

	###############################################################
	def derive(self, inputs):
		# Break it apart
		labels, label_lengths = inputs
		outputs = numpy.array([
			x
			for label, label_length in zip(labels, label_lengths)
			for x in label[:label_length[0]]
		])
		# Warp-CTC is strange: it uses a variable-length vector of
		# flattened labels. That doesn't look much like a deep
		# learning tensor! So we can cheat without imposing much of
		# a memory hit.
		outputs = numpy.lib.stride_tricks.as_strided(
			outputs,
			shape=(len(labels), ) + outputs.shape,
			strides=(0, ) + outputs.strides
		)
		return outputs

	###############################################################
	def shape(self):
		return (None,)

	###############################################################
	def requires(self):
		return (self.label, self.label_length)

###############################################################################
class Ctc(Loss):
	""" Connectionist Temporal Classification loss function
	"""

	###########################################################################
	def __init__(self, input_length, output_length, output, relative_to=None,
		variant=None, **kwargs):
		""" Creates a new CTC loss function.

			# Arguments
		"""
		super().__init__(**kwargs)

		if variant is None:
			self.variant = None
		elif variant == 'warp':
			self.variant = 'warp'
			if not can_import('ctc'):
				logger.error('The warp-CTC loss function was requested, but '
					'we cannot find the "ctc" library. See our '
					'troubleshooting page for helpful tips.')
				raise ImportError('Cannot find the "ctc" library, which is '
					'needed when using the "warp" variant of the CTC loss '
					'function.')
		else:
			raise ValueError('Unsupported CTC variant: {}'.format(variant))

		self.input_length = input_length
		self.output_length = output_length
		self.output = output
		self.relative_to = relative_to

	###########################################################################
	def get_loss(self, model, target, output):
		""" Returns the loss function that can be used by the implementation-
			specific model.
		"""
		backend = model.get_backend()

		if backend.get_name() == 'keras':

			import keras.backend as K

			if self.variant is None:

				# Just use the built-in Keras CTC loss function.
				logger.debug('Attaching built-in Keras CTC loss function to '
					'model output "%s".', target)

			elif self.variant == 'warp':

				# Just use the built-in Keras CTC loss function.
				logger.info('Attaching Warp-CTC loss function to model '
					'output "%s".', target)

				if backend.get_toolchain() != 'theano':
					logger.error('If you want to use warp-ctc, you need to '
						'use the Theano backend to Keras.')
					raise ValueError('Warp-CTC is currently only supported '
						'with the Theano backend to Keras.')

			else:
				raise ValueError('Unsupported variant "{}" on loss function '
					'"{}" for backend "{}".'.format(self.variant,
						self.get_name(), backend.get_name()))

			ctc_scaled = 'ctc_scaled_{}'.format(self.input_length)
			flattened_labels = 'ctc_flattened_labels_{}'.format(target)

			transcript_length = K.placeholder(
				ndim=2,
				dtype='int32',
				name=self.output_length
			)
			transcript = K.placeholder(
				ndim=2,
				dtype='int32',
				name=self.output if self.variant is None \
					else flattened_labels
			)
			utterance_length = K.placeholder(
				ndim=2,
				dtype='int32',
				name=self.input_length if self.relative_to is None \
					else ctc_scaled
			)

			if self.relative_to is not None:
				model.add_data_source(
					ctc_scaled,
					ScaledSource(
						model,
						relative_to=self.relative_to,
						to_this=target,
						scale_this=self.input_length
					)
				)

			if self.variant == 'warp':
				model.add_data_source(
					flattened_labels,
					FlattenSource(
						self.output,
						self.output_length
					)
				)

			if self.variant is None:
				out = K.ctc_batch_cost(
					transcript,
					output,
					utterance_length,
					transcript_length
				)
			else:
				import ctc						# pylint: disable=import-error
				out = ctc.cpu_ctc_th(
					output.dimshuffle((1, 0, 2)),
					K.squeeze(utterance_length, -1),
					transcript[0]+1,
					K.squeeze(transcript_length, -1)
				)

			return (
				(
					(self.output_length, transcript_length),
					(self.output if self.variant is None \
						else flattened_labels, transcript),
					(self.input_length if self.relative_to is None \
						else ctc_scaled, utterance_length)
				),
				out
			)

		else:
			raise ValueError('Unsupported backend "{}" for loss function "{}"'
				.format(backend.get_name(), self.get_name()))

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2016 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import logging
from . import Operator

logger = logging.getLogger(__name__)

###############################################################################
class Assertion(Operator):			# pylint: disable=too-few-public-methods
	""" Simple assertion layer for debugging purposes.
	"""

	###########################################################################
	@classmethod
	def get_container_name(cls):
		""" Returns the name of the container class.
		"""
		return 'assert'

	###########################################################################
	def __init__(self, *args, **kwargs):
		""" Create a new assertion container.
		"""
		super().__init__(*args, **kwargs)
		self.condition = None

	###########################################################################
	def _parse(self, engine):
		""" Parse the debug statement and print it.
		"""
		self.condition = engine.evaluate(self.args, recursive=True)
		if not self.condition:
			raise AssertionError('"assert" layer failed. Condition evaluated '
				'to: {} (type={})'.format(
					self.condition,
					type(self.condition)
				))

	###########################################################################
	def _build(self, model):	# pylint: disable=unused-argument,no-self-use
		""" Assert statements don't produce layers.
		"""
		return iter(())

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2016 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import json
from . import Supplier
from ..sources import JSONLSource

###############################################################################
class JSONLSupplier(Supplier):
    """ A supplier which parses out data from a JSON-Lines file (*.jsonl)
        the keys from the first JSON line object determine the source names, and
        all subsequent lines should have the same keys. Because JSON supports arrays,
        the JSONL Supplier is good for arbitrary tensor shapes, and scales to lots of
        data.
    """

    ###########################################################################
    @classmethod
    def get_name(cls):
        """ Returns the name of the supplier.
        """
        return 'jsonl'

    ###########################################################################
    def __init__(self, source, *args, **kwargs):
        """ Creates a new JSONL dictionary supplier.

            # Arguments

            source: str. Filename of JSONL file (*.jsonl) to load.
        """
        super().__init__(*args, **kwargs)

        self.source = source
        self.data = None
        with open(self.source, 'r') as infile:
            keys = json.loads(next(infile)).keys()

        with open(self.source, 'r') as infile:
            self.num_entries = sum(1 for _ in infile if _.strip())

        self.data = {
            k: JSONLSource(self.source, k, self.num_entries)
            for k in keys
        }

    ###########################################################################
    def get_sources(self, sources=None):
        """ Returns all sources from this provider.
        """

        if sources is None:
            with open(self.source, 'r') as infile:
                sources = list(json.loads(next(infile)).keys())

        elif not isinstance(sources, (list, tuple)):
            sources = [sources]

        for source in sources:
            if source not in self.data:
                raise KeyError(
                    'Invalid data key: {}. Valid keys are: {}'.format(
                        source, ', '.join(str(k) for k in self.data.keys())
                ))

        return {k : self.data[k] for k in sources}

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2016 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from .. import ParsingError
from .layer import Layer
from .placeholder import Placeholder
from .parallel import Parallel
from .expand import Expand
from .convolution import Convolution
from .reuse import Reuse
from .flatten import Flatten
from .dense import Dense
from .activation import Activation
from .pooling import Pooling
from .recurrent import Recurrent
from .batchnorm import BatchNormalization
from .output import Output
from .transpose import Transpose
from .merge import Merge
from .dropout import Dropout
from .repeat import Repeat
from .embedding import Embedding

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2017 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import logging
import numpy
from . import EvaluationHook#, TrainingHook

logger = logging.getLogger(__name__)


###############################################################################
class TextHook(EvaluationHook):#, TrainingHook):
    """ Post-evaluation hook for text data.
    """

    ###########################################################################
    @classmethod
    def get_name(cls):
        """ Returns the name of the evaluation hook.
        """
        return 'text'

    ###########################################################################
    @staticmethod
    def argmax_decode(output, rev_vocab):
        """ output = matrix: timesteps * characters
        """
        x = numpy.argmax(output, axis=1)
        tokens = []
        for c in x:
            tokens.append(c)
        return ''.join([rev_vocab[i] for i in tokens])

    ###########################################################################
    def __init__(self, **kwargs):
        """ Creates a new text hook.
        """

        super().__init__(**kwargs)

    ###########################################################################
    def apply(self, current, original, model=None):
        """ Applies the hook to the data.
        """

        _raw_input = {
            k: ''.join(original[1]['raw_' + k][0])
            for k in model.inputs.keys()
        }

        data, truth = current

        raw_truth = {
            k: ''.join(truth['raw_' + k][0])
            for k in model.outputs.keys()
        }

        vocabs = {
            k: model.provider.sources[model.provider.keys.index(k)].vocab
            for k in model.outputs.keys()
        }

        prediction = {
            k: self.argmax_decode(
                data[k][0],
                vocabs[k]
            )
            for k in model.outputs.keys()
        }

        print('Input:\n{}'.format('\n'.join([
            '%s: "%s"' % (k, v)
            for k, v in _raw_input.items()
        ])), flush=True)
        print('Prediction:\n{}'.format('\n'.join([
            '%s: "%s"' % (k, v)
            for k, v in prediction.items()
        ])), flush=True)
        print('Truth:\n{}'.format('\n'.join([
            '%s: "%s"' % (k, v)
            for k, v in raw_truth.items()
        ])), flush=True)
        return (prediction, raw_truth)

    # ###########################################################################
    # def notify(self, status, log=None, info=None):
    #     """ do not notify.
    #     """
    #     pass


#### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2017 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import functools
import logging
import os

import numpy

from .. import __homepage__

logger = logging.getLogger(__name__)

###############################################################################
class DependencyError(Exception):
	""" Exception class for missing dependencies.
	"""
	pass

###############################################################################
def load_wav(filename):
	""" Loads a WAV file.
	"""
	import scipy.io.wavfile as wav				# pylint: disable=import-error
	rate, sig = wav.read(filename)

	# Flatten stereo to mono
	if len(sig.shape) > 1:
		sig = numpy.array(sig.mean(axis=1), dtype=sig.dtype)

	return {
		'signal' : sig,
		'sample_rate' : rate,
		'sample_width' : sig.dtype.itemsize * 8,
		'channels' : 1
	}

###############################################################################
def load_pydub(filename):
	""" Loads an MP3 or FLAC file.
	"""
	try:
		from pydub import AudioSegment
		data = AudioSegment.from_file(filename)
	except ImportError:
		logger.exception('"pydub" is a required Python dependency for '
			'handling this audio file: %s.', filename)
		raise
	except FileNotFoundError:
		if os.path.isfile(filename):
			raise DependencyError()
		else:
			raise

	if data.channels > 1:
		data = functools.reduce(
			lambda x, y: x.overlay(y),
			data.split_to_mono()
		)

	raw = data.get_array_of_samples()
	raw = numpy.frombuffer(raw, dtype=raw.typecode)

	return {
		'signal' : raw,
		'sample_rate' : data.frame_rate,
		'sample_width' : data.sample_width * 8,
		'channels' : data.channels
	}

###############################################################################
def scale_signal(audio):
	""" Returns an audio signal scaled so that the max/min values are +1/-1.
	"""
	return audio['signal'] / 2**(audio['sample_width'] - 1)

###############################################################################
def get_mime_type(filename):
	""" Returns the MIME type associated with a particular audio file.
	"""
	try:
		import magic
	except ImportError:
		if get_mime_type.warn:
			logger.warning('Python package "magic" could not be loaded, '
				'possibly because system library "libmagic" could not be '
				'found. We are falling back on our own heuristics.')
			get_mime_type.warn = False

		ext = os.path.splitext(filename)[1].lower()
		return {
			'.wav' : 'audio/x-wav',
			'.mp3' : 'audio/mpeg',
			'.flac' : 'audio/x-flac'
		}.get(ext, 'unknown')
	else:
		# Read off magic numbers and return MIME types
		mime_magic = magic.Magic(mime=True)
		ftype = mime_magic.from_file(filename)
		if isinstance(ftype, bytes):
			ftype = ftype.decode('utf-8')

		# If we are dealing with a symlink, read the link
		# and try again with the target file.  We do this in 
		# a while loop to cover the case of symlinks which
		# point to other symlinks
		current_filename = filename
		while ftype == 'inode/symlink':
			current_filename = os.readlink(current_filename)
			ftype = mime_magic.from_file(current_filename)
			ftype = ftype.decode('utf-8') if isinstance(ftype, bytes) else ftype
			
		return ftype

get_mime_type.warn = True

###############################################################################
def load_audio(filename):
	""" Loads an audio file.

		# Arguments

		filename: str. The file to load.

		# Return value

		A dictionary with the following keys/values:
			- signal: A numpy array of raw audio data
			- sample_rate: The sample rate, in Hz
			- sample_width: The sample width, in bits
			- channels: The number of channels (always 1--see below)

		# Notes

		The returned audio data is mono. If the source file was stereo, it is
		downmixed to mono before being returned.
	"""
	ftype = get_mime_type(filename)

	# Define the loaders for each supported file type.
	loaders = {
		'audio/x-wav' : load_wav,
		'audio/mpeg' : load_pydub,
		'audio/x-flac' : load_pydub
	}

	if ftype not in loaders:
		raise IOError('No loader available for filetype ({}) for file: {}'
			.format(ftype, filename))

	try:
		return loaders[ftype](filename)
	except DependencyError:
		logger.exception('Failed to load audio file: %s. Its MIME type is: '
			'%s. The most likely cause is that you do not have FFMPEG '
			'installed. Check out our troubleshooting guide for more '
			'information: %s', filename, ftype,
			os.path.join(
				__homepage__,
				'troubleshooting.html#couldn-t-find-ffmpeg-or-avconv'
			)
		)
		raise
	except KeyboardInterrupt:
		raise
	except:
		logger.exception('Failed to load audio file: %s. Its MIME type is: '
			'%s. Check out our troubleshooting guide for more information: %s',
			filename, ftype, os.path.join(__homepage__, 'troubleshooting.html')
		)
		raise

###############################################################################
def get_audio_features(audio, feature_type, on_error=None, **kwargs):
	""" Returns audio features.

		# Arguments

		audio: dict or str. If dict, it should have keys, values as returned by
			`load_audio()`. If str, it should be a file that will be passed to
			`load_audio()`.
		feature_type. str. One of:
			- raw: Returns raw audio data (1-dimensional)
			- mfcc: Returns MFCC features
			- spec: Returns a spectrogram
		on_error: str or None (default: None). One of:
			- 'raise' or None: let the error propagate (no special catching)
			- 'suppress': catch the error and return None
		kwargs. Additional arguments that depend on `feature_type`:
			- For 'raw': no additional parameters
			- For 'mfcc':
				- features: int (default: 13). The number of MFCC features to
				  keep.
				- low_freq: int (default: None). The low-frequency cutoff.
				- high_freq: int (default: None). The high-frequency cutoff.
			- For 'spec':
				- low_freq: int (default: None). The low-frequency cutoff.
				- high_freq: int (default: None). The high-frequency cutoff.
	"""
	assert on_error in (None, 'suppress', 'raise')

	if isinstance(audio, str):
		original_path = audio
		audio = load_audio(audio)
	else:
		original_path = None

	if len(audio['signal']) < 1:
		logger.error('Failed to produce audio features while processing file '
			'%s. Length: %d. Sample rate: %d.', original_path,
			len(audio['signal']), audio['sample_rate'])
		if on_error == 'suppress':
			return None
		else:
			raise ValueError('Audio data is too short.')

	if feature_type == 'raw':
		return audio['signal']

	elif feature_type == 'mfcc':
		try:
			import python_speech_features
		except ImportError:
			logger.exception('"python_speech_features" is a required Python '
				'dependency for calculating MFCC features.')
			raise

		num_features = kwargs.get('features') or 13
		return python_speech_features.mfcc(
			audio['signal'],
			audio['sample_rate'],
			numcep=num_features,
			nfilt=num_features*2,
			lowfreq=kwargs.get('low_freq') or 0,
			highfreq=kwargs.get('high_freq') or None
		)

	elif feature_type == 'spec':
		# Window size, in seconds
		window_size = 0.020
		# Step size, in seconds
		step_size = 0.010

		signal = scale_signal(audio)

		hop_size = int(step_size * audio['sample_rate'])
		frame_size = int(window_size * audio['sample_rate'])

		if len(signal) < frame_size:
			logger.error('Failed to produce FFT while processing file '
				'%s. Original length: %d. Hop size: %d. Frame size: %d. '
				'Sample rate: %d.', original_path, len(signal), hop_size,
				frame_size, audio['sample_rate'])
			if on_error == 'suppress':
				return None
			else:
				raise ValueError('Audio data is too short.')

		# Cleave off any samples that do not cleanly fit into our step size.
		remove = (len(signal) - frame_size) % hop_size
		if remove:
			clean = signal[:-remove]
		else:
			clean = signal

		# Optimization: instead of doing a for loop or list comprehension to
		# apply the window to the signal, we can just create a new view into
		# the data with each window.
		num_frames = (len(clean) - frame_size) // hop_size + 1
		frames = numpy.lib.stride_tricks.as_strided(
			clean,
			shape=(frame_size, num_frames),
			strides=(clean.strides[0], clean.strides[0] * hop_size)
		)

		filter_window = numpy.hanning(frame_size)
		fft = numpy.fft.rfft(
			frames * numpy.expand_dims(filter_window, -1),
			axis=0
		)
		norm = numpy.absolute(fft)**2

		scale = numpy.sum(filter_window**2) * audio['sample_rate']
		scaled = norm
		scaled[1:-1] /= scale/2
		scaled[[0, -1]] /= scale

		spec = scaled
		# At this point, `spec` is shape (frequency, time).

		# Apply frequency cutoffs, if necessary
		low_freq = kwargs.get('low_freq')
		high_freq = kwargs.get('high_freq')
		if low_freq or high_freq:
			# Number of frequency bins
			num_bins = spec.shape[0]
			# Width of each frequency bin.
			delta_freq = 1 / window_size

			# Calculate the bin that a frequency would fall into.
			get_bin = lambda f, alt: \
				(
					min(
						max(int(f / delta_freq + 0.5), 0) + 1, num_bins
					)
					if f else alt
				)
			spec = spec[get_bin(low_freq, 0):get_bin(high_freq, num_bins)]

		spec = numpy.log(spec + 1e-14)
		# Format `spec` as (time, frequency)
		spec = spec.T
		return spec

	else:
		raise ValueError('Unsupported feature type: {}'.format(feature_type))

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2016 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import os
import gzip
import tarfile
import logging
import tempfile

from .network import get_hash, download

logger = logging.getLogger(__name__)

###############################################################################
def canonicalize(path):
	""" Returns an absolute, canonical path.
	"""
	return os.path.realpath(os.path.abspath(path))

###############################################################################
def sandbox_extract(path, dest, allow_outside_links=True,
	allow_absolute=False):
	""" Safely extract a tar archive, avoiding paths with try to "escape" from
		the destination path (the "sandbox"), since tar archives can accept
		files with absolute paths (e.g, "/") or relative paths ("../..", "..").

		# Arguments

		path: str. The path to the tarball to extract.
		dest: str. The destination folder to extract into.
		allow_outside_links: bool (default: True). If True, links inside the
			archive are allowed to point to files outside of the archive. If
			False, all links must resolve to files/directories in the archive.
		allow_absolute: bool (default: False). If True, files/links are allowed
			to have absolute paths; if False, all paths in the archive must be
			relative. In either case, any absolute paths must ultimately "end
			up" back inside the sandbox.

		# Return value

		A list of extracted filenames.
	"""

	def path_is_safe(path):
		""" We will paths that leave the current directory tree.
		"""
		if allow_absolute:
			# We allow absolute filenames, as long as they land inside our
			# destination directory.
			path = canonicalize(path)
			normalized_dest = canonicalize(dest)
			if not path.startswith(normalized_dest):
				return False
			# We could still have this problem:
			#   path = '/asd/fg__ha_ha_i_escaped/try_and_stop_me'
			#   normalized_dest = '/asd/fg'
			# So we need to demand that this falls on a directory boundary.
			suffix = path[len(normalized_dest):]
			return suffix.startswith('/')
		else:
			# We will simply forbid absolute paths altogether, as well as
			# relative paths that escape from our destination directory.
			path = os.path.normpath(path)
			if path.startswith('/') or path == '..' or path.startswith('../'):
				return False
			return True

	def link_is_safe(member):
		""" For links, they must resolve inside the current directory tree.
			Note: you still need to check that `path_is_safe(member.name)`.
		"""
		if not member.issym() and not member.islnk():
			return True
		return path_is_safe(
			os.path.join(os.path.dirname(member.name), member.linkname)
		)

	with tarfile.open(path) as tar:
		# Get a list of members that can be safely extracted.
		members = [
			member for member in tar.getmembers()
			if path_is_safe(member.name) and
				(allow_outside_links or link_is_safe(member))
		]
		# Extract the safe members
		tar.extractall(path=dest, members=members)
		# Return the list of file paths.
		return [canonicalize(os.path.join(dest, member.name))
			for member in members]

###############################################################################
def is_gzipped(filename):
	""" Returns True if the target filename looks like a GZIP'd file.
	"""
	with open(filename, 'rb') as fh:
		return fh.read(2) == b'\x1f\x8b'

###############################################################################
def gzip_extract(path, dest):
	""" Extracts a GZIP'd file into a target directory.
	"""
	with gzip.open(path, 'rb') as fh:
		content = fh.read()

	target = os.path.join(dest, os.path.basename(path))
	if path.endswith('.gz'):
		target = target[:-3]
	else:
		parts = os.path.splitext(target)
		if parts[1]:
			target = parts[0]
		else:
			target = target + '.content'

	with open(target, 'wb') as fh:
		fh.write(content)

	return canonicalize(target)

###############################################################################
def unpack(path, dest=None, recursive=False, ignore_error=False):
	""" Extracts an archive.

		# Arguments

		path: str. Filename to archive to extract.
		recursive: bool (default: False). If False, only the file at `path` is
			extracted. If True, then if any extracted files are themselves
			compressed, they will be recursively extracted as well.
		ignore_error: bool (default: False). If True, errors that result from
			trying to extract uncompressed files will be silently ignored.
		dest: str or None (default: None). Directory to extract to (will be
			created if it doesn't exist). If None, defaults to the same
			directory as `path`.
	"""
	if dest is None:
		dest = os.path.dirname(path)
	if not os.path.isdir(dest):
		if os.path.exists(dest):
			raise ValueError('Cannot extract to target location: {}'.format(
				dest))
		os.makedirs(dest, exist_ok=True)

	if os.path.isfile(path) and tarfile.is_tarfile(path):
		extracted = sandbox_extract(path, dest)
	elif os.path.isfile(path) and is_gzipped(path):
		extracted = gzip_extract(path, dest)
	else:
		if ignore_error:
			extracted = []
		else:
			raise ValueError('Unknown or unsupported file type: {}'
				.format(path))

	if recursive:
		for filename in list(extracted):
			extracted.extend(unpack(
				filename,
				os.path.dirname(filename),
				recursive=True,
				ignore_error=True
			))

	return extracted

###############################################################################
def install(url=None, path=None, checksum=None):
	""" Ensure that the data source exists locally.

		# Return value

		A tuple `(path, is_packed)`, where `path` is the path to the target on
		the local machine, and `is_packed` is a boolean which indicates
		whether or not `path` appears to be a single file (packed) or a
		directory (unpacked, as if it were a tar archive that was already
		extracted).
	"""
	if url is None:
		# Expect a path to an existing source
		if path is None:
			raise ValueError('Either "url" or "path" needs to be '
				'specified in the data supplier.')
		path = os.path.expanduser(os.path.expandvars(path))
		if os.path.isfile(path):
			# Perfect. Checksum it.
			if checksum is not None:
				actual = get_hash(path)
				if actual.lower() != checksum.lower():
					raise ValueError('Input file "{}" failed its '
						'checksum.'.format(path))
			return path, True
		elif os.path.isdir(path):
			return path, False
		else:
			raise ValueError('"path" was specified in a data supplier, but '
				'the path does not exist. Check that the path is correct, or '
				'specify a URL to download data.')
	else:
		if path is None:
			# URL, but no path: use temporary directory as path.
			path = tempfile.gettempdir()
		else:
			path = os.path.expanduser(os.path.expandvars(path))

		if not os.path.exists(path):
			# Create the necessary directories and download the file.
			os.makedirs(path, exist_ok=True)

		if os.path.isdir(path):
			# It's a directory that exists. Let's look for the would-be
			# downloaded file.
			_, filename = os.path.split(url)
			path = os.path.join(path, filename)

		if os.path.isfile(path):
			# File already exists. Checksum it.
			if checksum is not None:
				if get_hash(path).lower() == checksum.lower():
					logger.debug('File exists and passed checksum: %s',
						path)
					return path, True
				else:
					# Checksum fails -> redownload
					logger.warning('Input file "%s" failed its checksum. '
						'Redownloading...', path)
			else:
				logger.debug('File exists, but there is no checksum: %s',
					path)
				return path, True

		# Need to download the file.
		download(url, path)
		if checksum is not None:
			if get_hash(path).lower() != checksum.lower():
				raise ValueError('Failed to download URL: {}. The '
					'integrity check failed.'.format(url))
			else:
				logger.debug('Downloaded file passed checksum: %s', path)
		else:
			logger.debug('Downloaded file, but there is not checksum: %s',
				path)
		return path, True

	raise ValueError('Unhandled download path. This is a bug.')

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2016 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import os
import hashlib
import logging
import mimetypes
import urllib.request
import urllib.parse
import uuid
import json
import io
from collections import namedtuple

import tqdm

logger = logging.getLogger(__name__)

###############################################################################
UploadFile = namedtuple('UploadFile', 'filename')
UploadFileData = namedtuple('UploadFileData', ('filename', 'data'))

###############################################################################
def get_hash(path):
	""" Returns the SHA256 hash of a file.

		# Arguments

		path: str. Path to the file to produce a hash for.

		# Return value

		64-character lowercase string of hexadecimal digits representing the
		32-byte SHA256 hash of the file content.
	"""
	with open(path, 'rb') as fh:
		data = fh.read()
	sha = hashlib.sha256()
	sha.update(data)
	return sha.hexdigest()

###############################################################################
def download(url, target):
	""" Downloads a URL.
	"""
	logger.debug('Trying to download: %s', url)
	response = urllib.request.urlopen(url)
	with open(target, 'wb') as fh:
		with tqdm.tqdm(
			total=int(response.info().get('Content-Length')),
			unit='bytes',
			unit_scale=True,
			desc='Downloading'
		) as pbar:
			while True:
				chunk = response.read(8192)
				if not chunk:
					break
				fh.write(chunk)
				pbar.update(len(chunk))

	logger.info('File downloaded: %s', target)

###############################################################################
def prepare_json(data):
	""" Prepares a JSON-encoded HTTP message.
	"""
	data = json.dumps(data)
	data = data.encode('utf-8')
	header = {
		'Content-type' : 'application/json'
	}
	return (data, header)

###############################################################################
def prepare_multipart(data):
	""" Prepares an HTTP multipart message.
	"""
	# This is where we'll store our message as we build it.
	buffer = io.BytesIO()

	# Create our boundary ID.
	boundary = uuid.uuid4().hex

	# Loop over all entries in data.
	for k, v in data.items():

		# Do we need to upload a file?
		if isinstance(v, (UploadFile, UploadFileData)):
			buffer.write(
				'--{}\r\n'
				'Content-Disposition: form-data; name="{}"; filename="{}"\r\n'
				'Content-Type: {}\r\n'
				'\r\n'
					.format(
						boundary,
						urllib.parse.quote(k),
						urllib.parse.quote(
							os.path.basename(v.filename)
						),
						mimetypes.guess_type(v.filename)[0] or
							'application/octet-stream'
					).encode('utf-8')
			)
			if isinstance(v, UploadFile):
				with open(v.filename, 'rb') as fh:
					buffer.write(fh.read())
			else:
				if not isinstance(v.data, bytes):
					buffer.write(v.data.encode('utf-8'))
				else:
					buffer.write(v.data)
			buffer.write(b'\r\n')

		# Or just upload a simple value?
		else:
			buffer.write(
				'--{}\r\n'
				'Content-Disposition: form-data; name="{}"\r\n'
				'\r\n'
				'{}\r\n'
					.format(boundary, urllib.parse.quote(k), v).encode('utf-8')
			)

	# Write out the footer.
	buffer.write('--{}--\r\n'.format(boundary).encode('utf-8'))

	# Return everything.
	data = buffer.getvalue()
	header = {
		'Content-type' : 'multipart/form-data; boundary={}'.format(boundary)
	}

	return (data, header)

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



import linecache
import numpy
import json

from ..sources import ChunkSource


###############################################################################
class JSONLSource(ChunkSource):
    """ Data source for tensors stored in JSONL format
    """

    ###########################################################################
    def __init__(self, source, key, num_entries, *args, **kwargs):
        """ Creates a new JSONL source for file named `source`.
        """

        super().__init__(*args, **kwargs)

        self.source = source
        self.num_entries = num_entries
        self.key = key
        self.indices = numpy.arange(len(self))

    ###########################################################################
    def __iter__(self):
        """ Return an iterator to the data. Get the value (tensor) for self.key
        from each object and yield batches of these tensors
        """
        start = 0
        while start < self.num_entries:
            end = min(self.num_entries, start + self.chunk_size)

            # linecache line numbering starts at 1
            batch = [
                json.loads(linecache.getline(self.source, i + 1).strip())[self.key]
                for i in self.indices[start:end]
            ]

            yield batch
            start = end

    ###########################################################################
    def __len__(self):
        """ Returns the total number of entries that this source can return, if
            known.
        """
        return self.num_entries

    ###########################################################################
    def shape(self):
        """ Return the shape of the tensor (excluding batch size) returned by
            this data source.
        """
        return numpy.array(json.loads(linecache.getline(self.source, 0 + 1))[self.key]).shape

    ###########################################################################
    def can_shuffle(self):
        """ This source can be shuffled.
        """
        return True

    ###########################################################################
    def shuffle(self, indices):
        """ Applies a permutation to the data.
        """
        if len(indices) > len(self):
            raise ValueError('Shuffleable was asked to apply permutation, but '
                'the permutation is longer than the length of the data set.')
        self.indices[:len(indices)] = self.indices[:len(indices)][indices]



"""
Copyright 2016 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import numpy as np
from vocab import *
import json
import os


if not os.path.exists('./data/'):
    os.mkdir('./data/')

def one_hot(v, ndim):
    v_one_hot = np.zeros(
        (len(v), ndim,)
    )
    for i in range(len(v)):
        v_one_hot[i][v[i]] = 1.0
    return v_one_hot

x = []
y = []

all_chars = []
for book in [
    'pride_and_prejudice.txt',
    'shakespeare.txt'
]:
    with open('books/%s' % book, 'r') as infile:
        chars = [
            c for c in ' '.join(infile.read().lower().split())
            if c in set(vocab)
        ]
        all_chars += [' ']
        all_chars += chars

all_chars = list(' '.join(''.join(all_chars).split()))
num_chars = len(all_chars)
with open('cleaned.txt', 'w') as outfile:
    outfile.write(''.join(all_chars))


x, y = [], []

data_portions = [
    ('train', 0.8),
    ('validate', 0.05),
    ('test', 0.05),
    ('evaluate', 0.05),
]

dev = False
if dev:
    # shrink data to make things go faster
    for i in range(len(data_portions)):
        data_portions[i] = (
            data_portions[i][0],
            data_portions[i][1] * 0.1
        )

max_i = sum([
    int(round(len(all_chars) * fraction))
    for name, fraction in data_portions
]) - seq_len

for i in range(max_i):

    in_char_seq = all_chars[i: i + seq_len]

    # one hot representation
    sample_x = np.zeros((len(in_char_seq), n_vocab,))
    for j, c in enumerate(in_char_seq):
        sample_x[j][char_to_int[c]] = 1
    x.append(sample_x)

    sample_y = np.zeros(n_vocab)
    sample_y[char_to_int[all_chars[i + seq_len]]] = 1
    y.append(sample_y)

x, y = np.array(x).astype('int32'), np.array(y).astype('int32')

start_i = 0
for name, fraction in data_portions:
    end_i = start_i + int(round(len(x) * fraction))
    print(start_i, end_i)
    x0 = x[start_i: end_i]
    y0 = y[start_i: end_i]

    print('dims:')
    print(x0.shape)
    print(y0.shape)

    start_i = end_i

    with open('data/%s.jsonl' % name, 'w') as outfile:
        for sample_x, sample_y in zip(x0, y0):
            outfile.write(json.dumps({
                'in_seq': sample_x.tolist(),
                'out_char': sample_y.tolist()
            }))
            outfile.write('\n')

    del x0, y0
